/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

/*
 * This subroutine is made based on SPLIB.
 *
 * R. Bramley and X. Wang.
 * SPLIB: A library of iterative methods for sparse linear system.
 * Technical report, Indiana University--Bloomington, 1995.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "lislib.h"
#include "lis_precon.h"

int lis_symbolic_fact(LIS_PRECON precon);
int lis_numerical_fact(LIS_PRECON precon);

int lis_precon_create_iluk(LIS_PRECON precon)
{
	int n,nnz,k,err;
	LIS_SCALAR r;
	LIS_MATRIX P;
	
	n   = precon->Amat->n;
	nnz = precon->Amat->nnz;

	k  = precon->options[LIS_OPTIONS_FILL];
	r = precon->params[LIS_PARAMS_RELAX];

	if( k<0 ) precon->options[LIS_OPTIONS_FILL] = 0;
	if( r<0.0 || r>1.0 ) precon->params[LIS_PARAMS_RELAX] = 0.0;

	err = lis_matrix_create(n,0,precon->Amat->comm,&P);
	if( err )
	{
		return err;
	}

	precon->Pmat = P;
	lis_symbolic_fact(precon);
	lis_numerical_fact(precon);

	return LIS_SUCCESS;
}

#define NEW_BLOCK 400
int lis_symbolic_fact(LIS_PRECON precon)
{
	int n,nnz,m;
	int levfill,head,lowct,prev,actlev,lm,lmk,ijlum,ujbeg,ujend;
	int i,j,k,jj,rowct,k1,k2,nzi,nzlu,levp1;
	int	*ijlu,*uptr;
	int *lastcol,*levels,*rowll;
	LIS_MATRIX A;

	A       = precon->Amat;
	n       = A->n;
	nnz     = A->nnz;
	levfill = precon->options[LIS_OPTIONS_FILL];

	ijlu    = (int *)lis_malloc( (nnz+1)*sizeof(int));  /* Added by Jang */
	uptr    = (int *)lis_malloc( n*sizeof(int) );
	lastcol = (int *)lis_malloc( n*sizeof(int) );
	levels  = (int *)lis_malloc( (nnz+1)*sizeof(int) );
	rowll   = (int *)lis_malloc( n*sizeof(int) );		/* Added by Jang */

	ijlu[0] = n+1;
	uptr[0] = n+1;

	nzlu    = n;
	levp1   = levfill+1;

	for(i=0;i<n;i++) lastcol[i]=0;

	for(i=0;i<n;i++)
	{
#ifdef USE_MPI
		lis_sort_iid(A->ptr[i],A->ptr[i+1]-1,A->index,A->g_index,A->value);
#else
		lis_sort_id(A->ptr[i],A->ptr[i+1]-1,A->index,A->value);
#endif
		levels[i] = 0;
		nzi       = A->ptr[i+1]-A->ptr[i];
		if( nzi>1 )
		{
			/* n     = 5
			 * nzi   = 4
			 * rowll = [ 3, 1, 4, -, -] 
			 */
			rowct = 0;
			for(j=A->ptr[i];j<A->ptr[i+1];j++)
			{
				jj          = A->index[j];
				lastcol[jj] = i+1;
				if( jj!=i )
				{
					levels[jj]   = 1;
					rowll[rowct] = jj;
					rowct++;
				}
			}
			/* 
			 * rowll = [ 1, 3, 4, -, -] 
			 */
			lis_sort_i(0,nzi-2,rowll);
			/* 
			 * rowll = [ -, 3, -, 4, 5] 
			 */
			head = rowll[0];
			k1   = n;
			for(j=nzi-2;j>=0;j--)
			{
				k2        = rowll[j];
				rowll[k2] = k1;
				k1        = k2;
			}

			nzlu += nzi-1;
			j = head;
			lowct = 0;
			while( j<i )
			{
				lowct++;
				if( levfill!=0 )
				{
					ujbeg = uptr[j];
					ujend = ijlu[j+1];
					prev  = j;
					lm    = rowll[j];
					lmk   = levels[j];
					for(m=ujbeg;m<ujend;m++)
					{
						ijlum = ijlu[m];
						actlev = lmk + levels[m];
						if( lastcol[ijlum]!=i+1 )
						{
							if( actlev<=levp1 )
							{
								lastcol[ijlum] = i+1;
								levels[ijlum]  = actlev;
								while( lm<=ijlum )
								{
									prev = lm;
									lm   = rowll[lm];
								}
								rowll[prev] = ijlum;
								rowll[ijlum] = lm;
								prev = ijlum;
								nzlu++;
							}
						}
						else
						{
							levels[ijlum] = _min(levels[ijlum],actlev);
						}
					}
				}
				j = rowll[j];
			}
			if( nzlu>=nnz )
			{
				ijlu   = (int *)realloc(ijlu  , (nzlu + NEW_BLOCK)*sizeof(int)); /* Added by Jang. */
				levels = (int *)realloc(levels, (nzlu + NEW_BLOCK)*sizeof(int)); /* Added by Jang. */
				nnz  = nzlu + NEW_BLOCK;
			}
			ijlu[i+1] = nzlu+1;
			uptr[i] = ijlu[i] + lowct;
			j = head;
			k1 = ijlu[i];
			for(k=k1;k<=nzlu;k++)
			{
				ijlu[k] = j;
				levels[k] = levels[j];
				j = rowll[j];
			}
		}
		else
		{
			ijlu[i+1] = nzlu + 1;
			uptr[i] = ijlu[i];
		}
	}
	lis_free(levels);
	lis_free(rowll);
	lis_free(lastcol);
	precon->Pmat->index = ijlu;
	precon->Pmat->ptr   = uptr;
	precon->Pmat->nnz   = nnz;

	return LIS_SUCCESS;
}

int lis_numerical_fact(LIS_PRECON precon)
{
	int n,j,k,m,jj,mm,nrw,milu,nnz;
	int *colptrs;
	int *jlu, *uptr;
	LIS_SCALAR *lu;
	LIS_SCALAR rwnrm,mult,relax;
	LIS_MATRIX A;

	A       = precon->Amat;
	jlu     = precon->Pmat->index;
	uptr    = precon->Pmat->ptr;
	nnz     = precon->Pmat->nnz;
	relax   = precon->params[LIS_PARAMS_RELAX];
	n       = A->n;
	milu    = (relax!=0.0);

	colptrs = (int *)lis_malloc( n*sizeof(int) );						/* Added by Jang. */
	lu      = (LIS_SCALAR *)lis_malloc( (nnz+1)*sizeof(LIS_SCALAR) );	/* Added by Jang. */

	for(k=0;k<n;k++) colptrs[k] = 0;

	for(k=0;k<n;k++)
	{
		for(j=jlu[k];j<jlu[k+1];j++)
		{
			colptrs[jlu[j]] = j+1;
			lu[j]           = 0.0;
		}

		colptrs[k] = k+1;
		rwnrm = 0.0;
		nrw   = 0;
		for(j=A->ptr[k];j<A->ptr[k+1];j++)
		{
			jj = A->index[j];
			if( colptrs[jj]!=0 )
			{
				nrw++;
				rwnrm += fabs(A->value[j]);
				lu[colptrs[jj]-1] = A->value[j];
			}
		}

		for(j=jlu[k];j<uptr[k];j++)
		{
			jj = jlu[j];
			lu[j] = lu[j]*lu[jj];
			mult  = lu[j];
			if( milu )
			{
				for(m=uptr[jj];m<jlu[jj+1];m++)
				{
					mm = jlu[m];
					if( colptrs[mm]!=0 )
					{
						lu[colptrs[mm]-1] -= mult*lu[m];
					}
					else
					{
						lu[k] -= relax*mult*lu[m];
					}
				}
			}
			else
			{
				for(m=uptr[jj];m<jlu[jj+1];m++)
				{
					mm = jlu[m];
					if( colptrs[mm]!=0 )
					{
						lu[colptrs[mm]-1] -= mult*lu[m];
					}
				}
			}
		}

		for(j=jlu[k];j<jlu[k+1];j++)
		{
			colptrs[jlu[j]] = 0;
		}
		colptrs[k] = 0;

		lu[k] = 1.0/lu[k];
	}

	lis_free(colptrs);
	precon->Pmat->value = lu;
	return LIS_SUCCESS;

}

#ifdef n0
void lis_psolve_iluk(LIS_PRECON *precon, LIS_SCALAR b[], LIS_SCALAR x[])
{
	int i,j,n;
	int n1,n2,k,jj,is,ie,threads;
	LIS_SCALAR t;

	/*
	 *  LUx = b
	 */

	n = precon->Pmat->n;
	threads = omp_get_max_threads();
#pragma omp parallel for private(i,j,jj,is,ie,n1,n2,k,t)
	for(k=0;k<threads;k++)
	{
		n1 = n/threads;
		n2 = n%threads;
		if( k<n2 )
		{
			ie = n1+1;
			is = ie*k;
		}
		else
		{
			ie = n1;
			is = ie*k + n2;
		}
		ie = ie+is;

		for(i=is; i<ie; i++)
		{
			t = b[i];
			for(j=precon->Pmat->index[i];j<precon->Pmat->ptr[i];j++)
			{
				jj = precon->Pmat->index[j];
				if( jj>=is && jj<=ie ) t -= precon->Pmat->value[j] * x[jj];
//				t -= precon->Pmat->value[j] * x[precon->Pmat->index[j]];
			}
			x[i] = t;
		}
		for(i=ie-1; i>=is; i--)
		{
			t = x[i];
			for(j=precon->Pmat->ptr[i];j<precon->Pmat->index[i+1];j++)
			{
				jj = precon->Pmat->index[j];
				if( jj>=is && jj<=ie ) t -= precon->Pmat->value[j] * x[jj];
//				t -= precon->Pmat->value[j] * x[precon->Pmat->index[j]];
			}
			x[i] = t * precon->Pmat->value[i];
		}
	}
}

void lis_psolvet_iluk(LIS_PRECON *precon, LIS_SCALAR b[], LIS_SCALAR x[])
{
	int i,j,n;
	LIS_SCALAR t;

	/*
	 *  (LU)'x = b
	 */

	n = precon->Pmat->n;
	for(i=0; i<n; i++)
	{
		x[i] = b[i];
	}
	for(i=0; i<n; i++)
	{
		x[i] = x[i] * precon->Pmat->value[i];
		for(j=precon->Pmat->ptr[i];j<precon->Pmat->index[i+1];j++)
		{
			x[precon->Pmat->index[j]] -= precon->Pmat->value[j] * x[i];
		}
	}
	for(i=n-1; i>=0; i--)
	{
		for(j=precon->Pmat->index[i];j<precon->Pmat->ptr[i];j++)
		{
			x[precon->Pmat->index[j]] -= precon->Pmat->value[j] * x[i];
		}
	}
}
#endif

void lis_psolve_iluk(LIS_PRECON precon, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,j,n;
	LIS_SCALAR t;
	LIS_SCALAR *b,*x;

	/*
	 *  LUx = b
	 */

	n = precon->Pmat->n;
	b = B->value;
	x = X->value;
	for(i=0; i<n; i++)
	{
		t = b[i];
		for(j=precon->Pmat->index[i];j<precon->Pmat->ptr[i];j++)
		{
			t -= precon->Pmat->value[j] * x[precon->Pmat->index[j]];
		}
		x[i] = t;
	}
	for(i=n-1; i>=0; i--)
	{
		t = x[i];
		for(j=precon->Pmat->ptr[i];j<precon->Pmat->index[i+1];j++)
		{
			t -= precon->Pmat->value[j] * x[precon->Pmat->index[j]];
		}
		x[i] = t * precon->Pmat->value[i];
	}
}

void lis_psolvet_iluk(LIS_PRECON precon, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,j,n;
	LIS_SCALAR *b,*x;

	/*
	 *  (LU)'x = b
	 */

	n = precon->Pmat->n;
	b = B->value;
	x = X->value;
	for(i=0; i<n; i++)
	{
		x[i] = b[i];
	}
	for(i=0; i<n; i++)
	{
		x[i] = x[i] * precon->Pmat->value[i];
		for(j=precon->Pmat->ptr[i];j<precon->Pmat->index[i+1];j++)
		{
			x[precon->Pmat->index[j]] -= precon->Pmat->value[j] * x[i];
		}
	}
	for(i=n-1; i>=0; i--)
	{
		for(j=precon->Pmat->index[i];j<precon->Pmat->ptr[i];j++)
		{
			x[precon->Pmat->index[j]] -= precon->Pmat->value[j] * x[i];
		}
	}
}

