/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include "lislib.h"
#include "lis_solver.h"

int lis_bicgstabl(LIS_VECTOR b, LIS_VECTOR x, LIS_VECTOR work[],
		   LIS_SCALAR params[], int options[],
		   LIS_MATRIX A, LIS_PRECON M, LIS_SCALAR status[])
{
	LIS_SCALAR alpha, beta, omega, rho0, rho1;
	LIS_SCALAR rnorm0, rnorm;
	LIS_SCALAR normx, normr;
	LIS_SCALAR bnrm2, nrm2;
	LIS_SCALAR tol;
	LIS_SCALAR delta, one, zero, onem;
	LIS_SCALAR nu;
	int iter,maxiter,n,np,l,i,j,output;
	int w_dim, z_dim;
	int rcmp, xpdt;
	int ione, izero;
	double times,ptimes;

	LIS_VECTOR rtld, bp,t,xp, *r,*u;
	LIS_SCALAR *tau, *gamma, *gamma1, *gamma2, *sigma;

	n       = A->n;
	w_dim   = n;
	np      = A->np + options[LIS_OPTIONS_PADDING]; 
	maxiter = options[LIS_OPTIONS_MAXITER];
	output  = options[LIS_OPTIONS_OUTPUT];
	l       = options[LIS_OPTIONS_ELL];
	z_dim   = l+1;
	tol     = params[LIS_PARAMS_RESID];

	rtld    = work[0];
	xp      = work[1];
	bp      = work[2];
	t       = work[3];
	r       = &work[4];
	u       = &work[l+1+4];

	tau     = (LIS_SCALAR *)lis_malloc( sizeof(LIS_SCALAR) * z_dim * (4+l+1) );
	gamma   = &tau[z_dim*z_dim];
	gamma1  = &gamma[z_dim];
	gamma2  = &gamma1[z_dim];
	sigma   = &gamma2[z_dim];

	one     = (LIS_SCALAR)1.0;
	onem    = (LIS_SCALAR)-1.0;
	zero    = (LIS_SCALAR)0.0;
	ione    = 1;
	izero   = 0;

	ptimes  = 0.0;

	// Initial Residual
	if( lis_get_initial_residual(A,NULL,x,b,t,NULL,tol,&bnrm2,status) )
	{
		lis_free(tau);
		return LIS_SUCCESS;
	}
	times = lis_wtime();
	lis_psolve(M,t,r[0]);
	ptimes += lis_wtime()-times;
	lis_vector_copy(r[0],rtld);
//	lis_psolve(M,r,r);

	lis_vector_copy(r[0],bp);
	lis_vector_copy(x,xp);
	lis_vector_set_all(0.0,u[0]);
	lis_vector_nrm2(r[0],&rnorm0);
	rnorm  = rnorm0;
	normx  = rnorm0;
	normr  = rnorm0;
	rcmp   = 0;
	xpdt   = 0;

	alpha = (LIS_SCALAR)0.0;
	omega = (LIS_SCALAR)1.0;
	rho0  = (LIS_SCALAR)1.0;
	delta = (LIS_SCALAR)1.0e-2;


	iter = 0;
	while( iter<=maxiter )
	{
		/*           */
		/* BiCG PART */
		/*           */

		/* rho0 = -w*rho0 */
		rho0 = -omega*rho0;

		for( j=0; j<l; j++)
		{
			iter++;
			/* rho1 = <rtld,r[j]> */
			lis_vector_dot(rtld,r[j],&rho1);

			/* test breakdown */
			if( rho1==0.0 )
			{
				status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_BREAKDOWN;
				status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
				status[LIS_STATUS_RESID] = nrm2;
				lis_free(tau);
				return LIS_BREAKDOWN;
			}

			/* beta = alpha * (rho1/rho0) */
			/* rho0 = rho1                */
			beta = alpha*(rho1/rho0);
			rho0 = rho1;

			/* u[i] = r[i] - beta*u[i] (i=0,j) */
			for( i=0; i<=j; i++)
			{
				lis_vector_xpay(r[i],-beta,u[i]);
			}

			/* u[j+1] = A    * u[j]   */
			/* u[j+1] = M^-1 * u[j+1] */
			lis_matvec(A,u[j],t);
			times = lis_wtime();
			lis_psolve(M, t, u[j+1]);
			ptimes += lis_wtime()-times;

			/* nu = <rtld, u[j+1]> */
			lis_vector_dot(rtld,u[j+1],&nu);

			/* test breakdown */
			if( nu==0.0 )
			{
				status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_BREAKDOWN;
				status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
				status[LIS_STATUS_RESID] = nrm2;
				lis_free(tau);
				return LIS_BREAKDOWN;
			}

			/* alpha = rho1 / nu */
			alpha = rho1 / nu;

			/* x = x + alpha*u[0] */
			lis_vector_axpy(alpha,u[0],x);

			/* r[i] = r[i] - alpha*u[i+1] (i=0,j) */
			for( i=0; i<=j; i++)
			{
				lis_vector_axpy(-alpha,u[i+1],r[i]);
			}

			/* convergence check */
			lis_vector_nrm2(r[0],&nrm2);
			nrm2 = nrm2 * bnrm2;

			if( iter%l!=0 )
			{
				if( output )
				{
					if( output & LIS_PRINT_MEM ) lis_set_residual_history(nrm2);
					if( output & LIS_PRINT_OUT ) printf("iter: %5d  residual = %e\n", iter, nrm2);
				}
			}

			if( tol >= nrm2 )
			{
				if( output )
				{
					if( output & LIS_PRINT_MEM ) lis_set_residual_history(nrm2);
					if( output & LIS_PRINT_OUT ) printf("iter: %5d  residual = %e\n", iter, nrm2);
				}	
				lis_vector_axpy(1.0,xp,x);
				status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_SUCCESS;
				status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
				status[LIS_STATUS_RESID] = nrm2;
				status[LIS_STATUS_PTIMES] = ptimes;
				lis_free(tau);
				return LIS_SUCCESS;
			}

			/* r[j+1] = A    * r[j]   */
			/* r[j+1] = M^-1 * r[j+1] */
			lis_matvec(A,r[j],t);
			times = lis_wtime();
			lis_psolve(M, t, r[j+1]);
			ptimes += lis_wtime()-times;

			lis_vector_nrm2(r[0],&rnorm);
			normx = _max(normx,rnorm);
			normr = _max(normr,rnorm);
		}



		/*         */
		/* MR PART */
		/*         */

		for(j=1;j<=l;j++)
		{
			for(i=1;i<=j-1;i++)
			{
				lis_vector_dot(r[j],r[i],&nu);
				nu               = nu / sigma[i];
				tau[i*z_dim + j] = nu;
				lis_vector_axpy(-nu,r[i],r[j]);
			}
			lis_vector_dot(r[j],r[j],&sigma[j]);
			lis_vector_dot(r[0],r[j],&nu);
			gamma1[j] = nu / sigma[j]; 
		}

		gamma[l] = gamma1[l];
		omega    = gamma[l];
		for(j=l-1;j>=1;j--)
		{
			nu = 0.0;
			for(i=j+1;i<=l;i++)
			{
				nu += tau[j*z_dim + i]*gamma[i];
			}
			gamma[j] = gamma1[j] - nu;
		}
		for(j=1;j<=l-1;j++)
		{
			nu = 0.0;
			for(i=j+1;i<=l-1;i++)
			{
				nu += tau[j*z_dim + i]*gamma[i+1];
			}
			gamma2[j] = gamma[j+1] + nu;
		}

		/* UPDATE */
		lis_vector_axpy( gamma[1] ,r[0],x);
		lis_vector_axpy(-gamma1[l],r[l],r[0]);
		lis_vector_axpy(-gamma[l] ,u[l],u[0]);
		for(j=1;j<=l-1;j++)
		{
			lis_vector_axpy(-gamma[j] ,u[j],u[0]);
			lis_vector_axpy( gamma2[j],r[j],x);
			lis_vector_axpy(-gamma1[j],r[j],r[0]);
		}
		lis_vector_nrm2(r[0],&nrm2);
		nrm2 = nrm2 * bnrm2;

		if( output )
		{
			if( output & LIS_PRINT_MEM ) lis_set_residual_history(nrm2);
			if( output & LIS_PRINT_OUT ) printf("iter: %5d  residual = %e\n", iter, nrm2);
		}

		if( tol >= nrm2 )
		{
			lis_vector_axpy(1.0,xp,x);
			status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_SUCCESS;
			status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
			status[LIS_STATUS_RESID] = nrm2;
			status[LIS_STATUS_PTIMES] = ptimes;
			lis_free(tau);
			return LIS_SUCCESS;
		}
//		iter+=l;
	}

	status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_MAXITER;
	status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
	status[LIS_STATUS_RESID] = nrm2;
	lis_free(tau);
	return LIS_MAXITER;
}
