/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <math.h>
#include <stdlib.h>
#include <malloc.h>
#include "lislib.h"
#include "lis_matvec.h"
#include "lis_solver.h"

int lis_jacobi(LIS_VECTOR B, LIS_VECTOR X, LIS_VECTOR work[],
		   LIS_SCALAR params[], int options[],
		   LIS_MATRIX A, LIS_PRECON M, LIS_SCALAR status[])
{
	LIS_SCALAR ux,lx;
	LIS_SCALAR bb;
	LIS_SCALAR nrm2;
	LIS_SCALAR tol;
	LIS_SCALAR *r,*xt,*d;
	LIS_SCALAR *b,*x;
	LIS_VECTOR R,XT;
	int iter,maxiter,n,output;
	int i,j;
	int *diag;

	n       = A->n;
	maxiter = options[LIS_OPTIONS_MAXITER];
	output  = options[LIS_OPTIONS_OUTPUT];
	tol     = params[LIS_PARAMS_RESID];

	R       = work[0];
	XT      = work[1];
	x       = X->value;
	b       = B->value;
	r       = work[0]->value;
	xt      = work[1]->value;
	d       = work[2]->value;

	diag    = (int *)lis_malloc( n*sizeof(int) );

	lis_vector_nrm2(B,&bb);
	bb      = 1.0 / bb;
#pragma omp parallel for private(i,j)
	for(i=0; i<n; i++)
	{
		for(j=A->ptr[i];j<A->ptr[i+1];j++)
		{
			if( i==A->index[j] )
			{
				diag[i] = j;
				d[i]    = 1.0 / A->value[j];
				break;
			}
		}
	}

	for( iter=1; iter<=maxiter; iter++ )
	{
		lis_vector_copy(X,XT);
#ifdef USE_MPI
		if( A->np>XT->np )
		{
			XT->value = (LIS_SCALAR *)lis_realloc(XT->value,A->np*sizeof(LIS_SCALAR));
			if( XT->value==NULL )
			{
				SETERR("out of memory\n");
				return LIS_OUT_OF_MEMORY;
			}
			xt = XT->value;
		}
		lis_send_recv(A->commtable,xt);
#endif
#pragma omp parallel for private(i,j,lx,ux)
		for(i=0;i<n;i++)
		{
			lx = b[i];
			ux = 0.0;
			for(j=A->ptr[i];j<diag[i];j++)
			{
				lx -= A->value[j] * xt[A->index[j]];
			}
			for(j=diag[i]+1;j<A->ptr[i+1];j++)
			{
				ux -= A->value[j] * xt[A->index[j]];
			}
			lx  += ux;
			r[i] = lx - A->value[diag[i]]*x[i];
			x[i] = lx * d[i];
		}

		/* convergence check */
		lis_vector_nrm2(R,&nrm2);
		nrm2 = nrm2 * bb;

		if( output )
		{
			if( output & LIS_PRINT_MEM ) lis_set_residual_history(nrm2);
			if( output & LIS_PRINT_OUT ) printf("iter: %5d  residual = %e\n", iter, nrm2);
		}

		if( tol >= nrm2 )
		{
			status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_SUCCESS;
			status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
			status[LIS_STATUS_RESID] = nrm2;
			status[LIS_STATUS_PTIMES] = 0.0;
			lis_free(diag);
			return LIS_SUCCESS;
		}
	}

	status[LIS_STATUS_RET]   = (LIS_SCALAR)LIS_MAXITER;
	status[LIS_STATUS_ITER]  = (LIS_SCALAR)iter;
	status[LIS_STATUS_RESID] = nrm2;
	status[LIS_STATUS_PTIMES] = 0.0;
	lis_free(diag);
	return LIS_MAXITER;
}

